@@ -49,7 +49,7 @@
 #endif
 
 #include "php_amqp.h"
-#include "amqp_connection.h"
+#include "amqp_connection_resource.h"
 
 #if PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION >= 3
 zend_object_handlers amqp_channel_object_handlers;
@@ -65,7 +65,7 @@ HashTable *amqp_channel_object_get_debug_info(zval *object, int *is_temp TSRMLS_
 
 	/* Keep the first number matching the number of entries in this table*/
 	ALLOC_HASHTABLE(debug_info);
-	ZEND_INIT_SYMTABLE_EX(debug_info, 3 + 1, 0);
+	ZEND_INIT_SYMTABLE_EX(debug_info, 4 + 1, 0);
 
 	/* Start adding values */
 	MAKE_STD_ZVAL(value);
@@ -89,19 +89,63 @@ HashTable *amqp_channel_object_get_debug_info(zval *object, int *is_temp TSRMLS_
 }
 #endif
 
+void php_amqp_close_channel(amqp_channel_object *channel TSRMLS_DC)
+{
+	amqp_connection_object *connection;
+
+	assert(channel != NULL);
+
+	/* Pull out and verify the connection */
+	connection = AMQP_GET_CONNECTION(channel);
+
+	if (connection != NULL) {
+        /* First, remove it from active channels table to prevent recursion in case of connection error */
+        php_amqp_connection_resource_unregister_channel(connection->connection_resource, channel->channel_id);
+	} else {
+	    channel->is_connected = '\0';
+	}
+
+	if (!channel->is_connected) {
+		/* Nothing to do more - channel was previously marked as closed, possibly, due to channel-level error */
+		return;
+	}
+
+	channel->is_connected = '\0';
+
+	if (connection->connection_resource && connection->connection_resource->is_connected) {
+		assert(connection->connection_resource != NULL);
+
+		amqp_channel_close(connection->connection_resource->connection_state, channel->channel_id, AMQP_REPLY_SUCCESS);
+
+		amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+		if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+			PHP_AMQP_INIT_ERROR_MESSAGE();
+
+			php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+			php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+			php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+			PHP_AMQP_DESTROY_ERROR_MESSAGE();
+			return;
+		}
+
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+	}
+}
+
+
 void amqp_channel_dtor(void *object TSRMLS_DC)
 {
 	amqp_channel_object *channel = (amqp_channel_object*)object;
-	amqp_connection_object *connection;
 
-	AMQP_ASSIGN_CONNECTION(connection, channel);
-
-	if (connection) {
-		remove_channel_from_connection(connection, channel);
+	if (channel->is_connected) {
+		php_amqp_close_channel(channel TSRMLS_CC);
 	}
 
-	/* Destroy the connection storage */
-	if (channel->connection) {
+	if (channel->connection != NULL) {
+		/* Destroy the connection storage */
 		zval_ptr_dtor(&channel->connection);
 	}
 
@@ -149,8 +193,6 @@ PHP_METHOD(amqp_channel_class, __construct)
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
 
-	amqp_rpc_reply_t res;
-
 	/* Parse out the method parameters */
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "OO", &id, amqp_channel_class_entry, &connection_object, amqp_connection_class_entry) == FAILURE) {
 		zend_throw_exception(amqp_channel_exception_class_entry, "Parameter must be an instance of AMQPConnection.", 0 TSRMLS_CC);
@@ -170,29 +212,42 @@ PHP_METHOD(amqp_channel_class, __construct)
 	AMQP_VERIFY_CONNECTION(connection, "Could not create channel.");
 
 	/* Figure out what the next available channel is on this connection */
-	channel->channel_id = get_next_available_channel(connection, channel);
+	channel->channel_id = php_amqp_connection_resource_get_available_channel_id(connection->connection_resource);
 
 	/* Check that we got a valid channel */
-	if (channel->channel_id < 0) {
+	if (!channel->channel_id) {
 		zend_throw_exception(amqp_channel_exception_class_entry, "Could not create channel. Connection has no open channel slots remaining.", 0 TSRMLS_CC);
 		return;
 	}
 
+	if (php_amqp_connection_resource_register_channel(connection->connection_resource, channel, channel->channel_id) == FAILURE) {
+		zend_throw_exception(amqp_channel_exception_class_entry, "Could not create channel. Failed to add channel to connection slot.", 0 TSRMLS_CC);
+	}
+
 	/* Open up the channel for use */
-	amqp_channel_open(connection->connection_resource->connection_state, channel->channel_id);
+	amqp_channel_open_ok_t *r = amqp_channel_open(connection->connection_resource->connection_state, channel->channel_id);
 
-	res = AMQP_RPC_REPLY_T_CAST amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char ** pstr = (char **) &str;
-		amqp_error(res, pstr, connection, channel);
+	if (!r) {
+		amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-		zend_throw_exception(amqp_channel_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
+
+		php_amqp_connection_resource_unregister_channel(connection->connection_resource, channel->channel_id);
+		channel->channel_id = 0;
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	/* r->channel_id is a 16-bit channel number insibe amqp_bytes_t, assertion below will without converting to uint16_t*/
+	/* assert (r->channel_id == channel->channel_id);*/
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	channel->is_connected = '\1';
 
@@ -202,8 +257,25 @@ PHP_METHOD(amqp_channel_class, __construct)
 		channel->channel_id,
 		0,							/* prefetch window size */
 		channel->prefetch_count,	/* prefetch message count */
+		/* NOTE that RabbitMQ has reinterpreted global flag field. See https://www.rabbitmq.com/amqp-0-9-1-reference.html#basic.qos.global for details */
 		0							/* global flag */
 	);
+
+	amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
+		return;
+	}
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 }
 /* }}} */
 
@@ -269,10 +341,6 @@ PHP_METHOD(amqp_channel_class, setPrefetchCount)
 	/* Get the channel object out of the store */
 	channel = (amqp_channel_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	/* Set the prefetch count - the implication is to disable the size */
-	channel->prefetch_count = prefetch_count;
-	channel->prefetch_size = 0;
-
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not set prefetch count.");
 
@@ -281,12 +349,32 @@ PHP_METHOD(amqp_channel_class, setPrefetchCount)
 		amqp_basic_qos(
 			connection->connection_resource->connection_state,
 			channel->channel_id,
-			channel->prefetch_size,
-			channel->prefetch_count,
+			0,
+			prefetch_count,
 			0
 		);
+
+		amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+		if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+			PHP_AMQP_INIT_ERROR_MESSAGE();
+
+			php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+			php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+			php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+			PHP_AMQP_DESTROY_ERROR_MESSAGE();
+			return;
+		}
+
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
 	}
 
+	/* Set the prefetch count - the implication is to disable the size */
+	channel->prefetch_count = prefetch_count;
+	channel->prefetch_size = 0;
+
 	RETURN_TRUE;
 }
 /* }}} */
@@ -325,10 +413,6 @@ PHP_METHOD(amqp_channel_class, setPrefetchSize)
 	/* Get the channel object out of the store */
 	channel = (amqp_channel_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-	/* Set the prefetch size - the implication is to disable the count */
-	channel->prefetch_count = 0;
-	channel->prefetch_size = prefetch_size;
-
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not set prefetch size.");
 
@@ -337,12 +421,32 @@ PHP_METHOD(amqp_channel_class, setPrefetchSize)
 		amqp_basic_qos(
 			connection->connection_resource->connection_state,
 			channel->channel_id,
-			channel->prefetch_size,
-			channel->prefetch_count,
+			prefetch_size,
+			0,
 			0
 		);
+
+		amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+		if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+			PHP_AMQP_INIT_ERROR_MESSAGE();
+
+			php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+			php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+			php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+			PHP_AMQP_DESTROY_ERROR_MESSAGE();
+			return;
+		}
+
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
 	}
 
+	/* Set the prefetch size - the implication is to disable the count */
+	channel->prefetch_count = 0;
+	channel->prefetch_size = prefetch_size;
+
 	RETURN_TRUE;
 }
 /* }}} */
@@ -398,8 +502,25 @@ PHP_METHOD(amqp_channel_class, qos)
 			channel->channel_id,
 			channel->prefetch_size,
 			channel->prefetch_count,
+			/* NOTE that RabbitMQ has reinterpreted global flag field. See https://www.rabbitmq.com/amqp-0-9-1-reference.html#basic.qos.global for details */
 			0							/* Global flag - whether this change should affect every channel */
 		);
+
+		amqp_rpc_reply_t res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+		if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+			PHP_AMQP_INIT_ERROR_MESSAGE();
+
+			php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+			php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+			php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+			PHP_AMQP_DESTROY_ERROR_MESSAGE();
+			return;
+		}
+
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
 	}
 
 	RETURN_TRUE;
@@ -414,10 +535,8 @@ PHP_METHOD(amqp_channel_class, startTransaction)
 	zval *id;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
-	amqp_tx_select_t s;
-	amqp_method_number_t select_ok = AMQP_TX_SELECT_OK_METHOD;
-	amqp_rpc_reply_t res;
 
+	amqp_rpc_reply_t res;
 
 	/* Get the vhost from the method params */
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O", &id, amqp_channel_class_entry) == FAILURE) {
@@ -430,25 +549,26 @@ PHP_METHOD(amqp_channel_class, startTransaction)
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not start the transaction.");
 
-	res = AMQP_RPC_REPLY_T_CAST amqp_simple_rpc(
+	amqp_tx_select(
 		connection->connection_resource->connection_state,
-		channel->channel_id,
-		AMQP_TX_SELECT_METHOD,
-		&select_ok,
-		&s
+		channel->channel_id
 	);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char **pstr = (char **)&str;
-		amqp_error(res, pstr, connection, channel);
+	res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-		channel->is_connected = 0;
-		zend_throw_exception(amqp_channel_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	RETURN_TRUE;
 }
@@ -462,10 +582,8 @@ PHP_METHOD(amqp_channel_class, commitTransaction)
 	zval *id;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
-	amqp_tx_commit_t s;
-	amqp_rpc_reply_t res;
-	amqp_method_number_t commit_ok = AMQP_TX_COMMIT_OK_METHOD;
 
+	amqp_rpc_reply_t res;
 
 	/* Get the vhost from the method params */
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O", &id, amqp_channel_class_entry) == FAILURE) {
@@ -478,25 +596,26 @@ PHP_METHOD(amqp_channel_class, commitTransaction)
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not start the transaction.");
 
-	res = AMQP_RPC_REPLY_T_CAST amqp_simple_rpc(
+	amqp_tx_commit(
 		connection->connection_resource->connection_state,
-		channel->channel_id,
-		AMQP_TX_COMMIT_METHOD,
-		&commit_ok,
-		&s
+		channel->channel_id
 	);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char **pstr = (char **)&str;
-		amqp_error(res, pstr, connection, channel);
+	res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-		channel->is_connected = 0;
-		zend_throw_exception(amqp_channel_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	RETURN_TRUE;
 }
@@ -509,10 +628,8 @@ PHP_METHOD(amqp_channel_class, rollbackTransaction)
 	zval *id;
 	amqp_channel_object *channel;
 	amqp_connection_object *connection;
-	amqp_tx_rollback_t s;
-	amqp_method_number_t rollback_ok = AMQP_TX_ROLLBACK_OK_METHOD;
-	amqp_rpc_reply_t res;
 
+	amqp_rpc_reply_t res;
 
 	/* Get the vhost from the method params */
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O", &id, amqp_channel_class_entry) == FAILURE) {
@@ -525,25 +642,26 @@ PHP_METHOD(amqp_channel_class, rollbackTransaction)
 	connection = AMQP_GET_CONNECTION(channel);
 	AMQP_VERIFY_CONNECTION(connection, "Could not start the transaction.");
 
-	res = AMQP_RPC_REPLY_T_CAST amqp_simple_rpc(
+	amqp_tx_rollback(
 		connection->connection_resource->connection_state,
-		channel->channel_id,
-		AMQP_TX_ROLLBACK_METHOD,
-		&rollback_ok,
-		&s
+		channel->channel_id
 	);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char str[256];
-		char **pstr = (char **)&str;
-		amqp_error(res, pstr, connection, channel);
+	res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
 
-		channel->is_connected = 0;
-		zend_throw_exception(amqp_channel_exception_class_entry, *pstr, 0 TSRMLS_CC);
-		amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
 		return;
 	}
-	amqp_maybe_release_buffers(connection->connection_resource->connection_state);
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
 
 	RETURN_TRUE;
 }
@@ -562,7 +680,57 @@ PHP_METHOD(amqp_channel_class, getConnection)
 
 	channel = (amqp_channel_object *)zend_object_store_get_object(id TSRMLS_CC);
 
-    RETURN_ZVAL(channel->connection, 1, 0);
+	RETURN_ZVAL(channel->connection, 1, 0);
+}
+/* }}} */
+
+/* {{{ proto bool amqp::basicRecover([bool requeue=TRUE])
+Redeliver unacknowledged messages */
+PHP_METHOD(amqp_channel_class, basicRecover)
+{
+	zval *id;
+	amqp_channel_object *channel;
+	amqp_connection_object *connection;
+
+	amqp_rpc_reply_t res;
+
+	zend_bool requeue = 1;
+
+	/* Get the vhost from the method params */
+	if (zend_parse_method_parameters(ZEND_NUM_ARGS() TSRMLS_CC, getThis(), "O|b", &id, amqp_channel_class_entry, &requeue) == FAILURE) {
+		return;
+	}
+
+	/* Get the channel object out of the store */
+	channel = (amqp_channel_object *)zend_object_store_get_object(id TSRMLS_CC);
+
+	connection = AMQP_GET_CONNECTION(channel);
+	AMQP_VERIFY_CONNECTION(connection, "Could not redeliver unacknowledged messages.");
+
+	amqp_basic_recover(
+		connection->connection_resource->connection_state,
+		channel->channel_id,
+		(amqp_boolean_t) requeue
+	);
+
+	res = amqp_get_rpc_reply(connection->connection_resource->connection_state);
+
+	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
+		PHP_AMQP_INIT_ERROR_MESSAGE();
+
+		php_amqp_error(res, PHP_AMQP_ERROR_MESSAGE_PTR, connection, channel TSRMLS_CC);
+
+		php_amqp_zend_throw_exception(res, amqp_channel_exception_class_entry, PHP_AMQP_ERROR_MESSAGE, 0 TSRMLS_CC);
+		php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+		PHP_AMQP_DESTROY_ERROR_MESSAGE();
+		return;
+	}
+
+	php_amqp_maybe_release_buffers_on_channel(connection, channel);
+
+	RETURN_TRUE;
+
 }
 /* }}} */
 
