@@ -23,7 +23,7 @@
 
 /* $Id: amqp_channel.h 305865 2010-12-01 01:30:56Z pdezwart $ */
 
-
+void php_amqp_close_channel(amqp_channel_object *channel TSRMLS_DC);
 void amqp_channel_dtor(void *object TSRMLS_DC);
 zend_object_value amqp_channel_ctor(zend_class_entry *ce TSRMLS_DC);
 
@@ -42,6 +42,8 @@ PHP_METHOD(amqp_channel_class, rollbackTransaction);
 
 PHP_METHOD(amqp_channel_class, getConnection);
 
+PHP_METHOD(amqp_channel_class, basicRecover);
+
 /*
 *Local variables:
 *tab-width: 4
